@@ -20,7 +20,7 @@
 #include "nsAccUtils.h"
 #include "DocAccessibleParent.h"
 #include "Relation.h"
-#include "Role.h"
+#include "mozilla/a11y/Role.h"
 #include "RootAccessible.h"
 #include "mozilla/a11y/PDocAccessible.h"
 #include "mozilla/dom/BrowserParent.h"
@@ -43,7 +43,6 @@ using namespace mozilla::a11y;
 @interface mozAccessible ()
 - (BOOL)providesLabelNotTitle;
 
-- (void)maybePostLiveRegionChanged;
 - (void)maybePostA11yUtilNotification;
 @end
 
@@ -123,6 +122,10 @@ using namespace mozilla::a11y;
   if (state == states::BUSY) {
     [self moxPostNotification:@"AXElementBusyChanged"];
   }
+
+  if (state == states::EXPANDED) {
+    [self moxPostNotification:@"AXExpandedChanged"];
+  }
 }
 
 - (BOOL)providesLabelNotTitle {
@@ -159,6 +162,12 @@ using namespace mozilla::a11y;
     return ![self moxIsLiveRegion];
   }
 
+  if (selector == @selector(moxARIAPosInSet) || selector == @selector
+                                                    (moxARIASetSize)) {
+    GroupPos groupPos = mGeckoAccessible->GroupPosition();
+    return groupPos.setSize == 0;
+  }
+
   if (selector == @selector(moxExpanded)) {
     return [self stateWithMask:states::EXPANDABLE] == 0;
   }
@@ -289,7 +298,8 @@ using namespace mozilla::a11y;
 
 - (NSString*)moxRole {
 #define ROLE(geckoRole, stringRole, ariaRole, atkRole, macRole, macSubrole, \
-             msaaRole, ia2Role, androidClass, nameRule)                     \
+             msaaRole, ia2Role, androidClass, iosIsElement, uiaControlType, \
+             nameRule)                                                      \
   case roles::geckoRole:                                                    \
     return macRole;
 
@@ -360,7 +370,8 @@ using namespace mozilla::a11y;
   }
 
 #define ROLE(geckoRole, stringRole, ariaRole, atkRole, macRole, macSubrole, \
-             msaaRole, ia2Role, androidClass, nameRule)                     \
+             msaaRole, ia2Role, androidClass, iosIsElement, uiaControlType, \
+             nameRule)                                                      \
   case roles::geckoRole:                                                    \
     if (![macSubrole isEqualToString:NSAccessibilityUnknownSubrole]) {      \
       return macSubrole;                                                    \
@@ -376,7 +387,7 @@ using namespace mozilla::a11y;
   // and are instructed by the ARIA map to use the native host role.
   roleAtom = [self ARIARole];
 
-  if (roleAtom == nsGkAtoms::log_) {
+  if (roleAtom == nsGkAtoms::log) {
     return @"AXApplicationLog";
   }
 
@@ -400,7 +411,7 @@ struct RoleDescrMap {
   const nsString description;
 };
 
-static const RoleDescrMap sRoleDescrMap[] = {
+MOZ_RUNINIT static const RoleDescrMap sRoleDescrMap[] = {
     {@"AXApplicationAlert", u"alert"_ns},
     {@"AXApplicationAlertDialog", u"alertDialog"_ns},
     {@"AXApplicationDialog", u"dialog"_ns},
@@ -457,7 +468,7 @@ struct RoleDescrComparator {
 
   if (subrole) {
     size_t idx = 0;
-    if (BinarySearchIf(sRoleDescrMap, 0, ArrayLength(sRoleDescrMap),
+    if (BinarySearchIf(sRoleDescrMap, 0, std::size(sRoleDescrMap),
                        RoleDescrComparator(subrole), &idx)) {
       return utils::LocalizedString(sRoleDescrMap[idx].description);
     }
@@ -575,6 +586,26 @@ struct RoleDescrComparator {
   return @YES;
 }
 
+- (NSString*)moxInvalid {
+  // For controls that support text input, we will expose
+  // the string value of `aria-invalid` when it exists.
+  // See mozTextAccessible::moxInvalid for that work.
+  // Unfortunately, NSBools do not autoconvert to usable
+  // NSStrings, so we expose "true" and "false" manually.
+  return ([self stateWithMask:states::INVALID] != 0) ? @"true" : @"false";
+}
+
+- (NSArray*)moxErrorMessageElements {
+  if (![[self moxInvalid] isEqualToString:@"false"]) {
+    NSArray* relations = [self getRelationsByType:RelationType::ERRORMSG];
+    if ([relations count] > 0) {
+      return relations;
+    }
+  }
+
+  return nil;
+}
+
 - (NSNumber*)moxFocused {
   return @([self stateWithMask:states::FOCUSED] != 0);
 }
@@ -620,6 +651,16 @@ struct RoleDescrComparator {
   return utils::GetAccAttr(self, nsGkAtoms::aria_live);
 }
 
+- (NSNumber*)moxARIAPosInSet {
+  GroupPos groupPos = mGeckoAccessible->GroupPosition();
+  return @(groupPos.posInSet);
+}
+
+- (NSNumber*)moxARIASetSize {
+  GroupPos groupPos = mGeckoAccessible->GroupPosition();
+  return @(groupPos.setSize);
+}
+
 - (NSString*)moxARIARelevant {
   if (NSString* relevant =
           utils::GetAccAttr(self, nsGkAtoms::containerRelevant)) {
@@ -630,6 +671,16 @@ struct RoleDescrComparator {
   return @"additions text";
 }
 
+- (NSString*)moxPlaceholderValue {
+  // First, check for plaecholder HTML attribute
+  if (NSString* placeholder = utils::GetAccAttr(self, nsGkAtoms::placeholder)) {
+    return placeholder;
+  }
+
+  // If no placeholder HTML attribute, check for the aria version.
+  return utils::GetAccAttr(self, nsGkAtoms::aria_placeholder);
+}
+
 - (id)moxTitleUIElement {
   MOZ_ASSERT(mGeckoAccessible);
 
@@ -729,10 +780,37 @@ struct RoleDescrComparator {
   return [self moxEditableAncestor];
 }
 
+- (NSString*)moxLanguage {
+  MOZ_ASSERT(mGeckoAccessible);
+
+  nsAutoString lang;
+  mGeckoAccessible->Language(lang);
+
+  return nsCocoaUtils::ToNSString(lang);
+}
+
+- (NSString*)moxKeyShortcutsValue {
+  MOZ_ASSERT(mGeckoAccessible);
+
+  nsAutoString shortcut;
+
+  if (!mGeckoAccessible->GetStringARIAAttr(nsGkAtoms::aria_keyshortcuts,
+                                           shortcut)) {
+    return nil;
+  }
+
+  return nsCocoaUtils::ToNSString(shortcut);
+}
+
 #ifndef RELEASE_OR_BETA
 - (NSString*)moxMozDebugDescription {
   NS_OBJC_BEGIN_TRY_BLOCK_RETURN;
 
+  if (!mGeckoAccessible) {
+    return [NSString stringWithFormat:@"<%@: %p mGeckoAccessible=null>",
+                                      NSStringFromClass([self class]), self];
+  }
+
   NSMutableString* domInfo = [NSMutableString string];
   if (NSString* tagName = utils::GetAccAttr(self, nsGkAtoms::tag)) {
     [domInfo appendFormat:@" %@", tagName];
@@ -823,17 +901,6 @@ struct RoleDescrComparator {
   return NO;
 }
 
-- (void)maybePostLiveRegionChanged {
-  id<MOXAccessible> liveRegion =
-      [self moxFindAncestor:^BOOL(id<MOXAccessible> moxAcc, BOOL* stop) {
-        return [moxAcc moxIsLiveRegion];
-      }];
-
-  if (liveRegion) {
-    [liveRegion moxPostNotification:@"AXLiveRegionChanged"];
-  }
-}
-
 - (void)maybePostA11yUtilNotification {
   MOZ_ASSERT(mGeckoAccessible);
   // Sometimes we use a special live region to make announcements to the user.
@@ -846,40 +913,43 @@ struct RoleDescrComparator {
   // a random acc with the same ID) by checking:
   //  - The gecko acc is local, our a11y-announcement lives in browser.xhtml
   //  - The ID of the gecko acc is "a11y-announcement"
-  //  - The native acc is a direct descendent of the root
-  if (mGeckoAccessible->IsLocal() &&
-      [[self moxDOMIdentifier] isEqualToString:@"a11y-announcement"] &&
-      [[self moxParent] isKindOfClass:[mozRootAccessible class]]) {
-    // Our actual announcement should be stored as a child of the alert,
-    // so we verify a child exists, and then query that child below.
-    NSArray* children = [self moxChildren];
-    MOZ_ASSERT([children count] == 1 && children[0],
-               "A11yUtil event recieved, but no announcement found?");
-
-    mozAccessible* announcement = children[0];
-    NSString* key;
-    if ([announcement providesLabelNotTitle]) {
-      key = [announcement moxLabel];
+  //  - The native acc is a direct descendent of the chrome window (ChildView in
+  //  a non-headless context, mozRootAccessible in a headless context).
+  DocAccessible* maybeRoot = mGeckoAccessible->IsLocal()
+                                 ? mGeckoAccessible->AsLocal()->Document()
+                                 : nullptr;
+  if (maybeRoot && maybeRoot->IsRoot() &&
+      [[self moxDOMIdentifier] isEqualToString:@"a11y-announcement"]) {
+    nsAutoString name;
+    // Our actual announcement should be stored as a child of the alert.
+    if (Accessible* announcement = mGeckoAccessible->FirstChild()) {
+      announcement->Name(name);
     } else {
-      key = [announcement moxTitle];
+      MOZ_ASSERT_UNREACHABLE(
+          "A11yUtil event received, but no announcement found?");
     }
 
     NSDictionary* info = @{
-      NSAccessibilityAnnouncementKey : key ? key : @(""),
-      NSAccessibilityPriorityKey : @(NSAccessibilityPriorityMedium)
+      NSAccessibilityAnnouncementKey : name.IsEmpty()
+          ? @("")
+          : nsCocoaUtils::ToNSString(name),
+      // High priority means VO will stop what it is currently speaking
+      // to speak our announcement.
+      NSAccessibilityPriorityKey : @(NSAccessibilityPriorityHigh)
     };
 
-    id window = [self moxWindow];
-
     // This sends events via nsIObserverService to be consumed by our
     // mochitests. Normally we'd fire these events through moxPostNotification
-    // which takes care of this, but because the window we fetch above isn't
-    // derrived from MOXAccessibleBase, we do this (and post the notification)
-    // manually.
+    // which takes care of this, but because NSApp isn't derived
+    // from MOXAccessibleBase, we do this (and post the notification) manually.
+    // We used to fire this on the window, but per Chrome and Safari these
+    // notifs get dropped if fired on any non-main window. We now fire on NSApp
+    // to avoid this.
     xpcAccessibleMacEvent::FireEvent(
-        window, NSAccessibilityAnnouncementRequestedNotification, info);
+        GetNativeFromGeckoAccessible(maybeRoot),
+        NSAccessibilityAnnouncementRequestedNotification, info);
     NSAccessibilityPostNotificationWithUserInfo(
-        window, NSAccessibilityAnnouncementRequestedNotification, info);
+        NSApp, NSAccessibilityAnnouncementRequestedNotification, info);
   }
 }
 
@@ -900,6 +970,7 @@ struct RoleDescrComparator {
                                inserted:(BOOL)isInserted
                             inContainer:(Accessible*)container
                                      at:(int32_t)start {
+  [self maybePostValidationErrorChanged];
 }
 
 - (void)handleAccessibleEvent:(uint32_t)eventType {
@@ -952,15 +1023,40 @@ struct RoleDescrComparator {
     case nsIAccessibleEvent::EVENT_LIVE_REGION_REMOVED:
       mIsLiveRegion = false;
       break;
-    case nsIAccessibleEvent::EVENT_REORDER:
-      [self maybePostLiveRegionChanged];
-      break;
-    case nsIAccessibleEvent::EVENT_NAME_CHANGE: {
+    case nsIAccessibleEvent::EVENT_NAME_CHANGE:
       if (![self providesLabelNotTitle]) {
         [self moxPostNotification:NSAccessibilityTitleChangedNotification];
       }
-      [self maybePostLiveRegionChanged];
       break;
+    case nsIAccessibleEvent::EVENT_LIVE_REGION_CHANGED:
+      MOZ_ASSERT(mIsLiveRegion);
+      [self moxPostNotification:@"AXLiveRegionChanged"];
+      break;
+    case nsIAccessibleEvent::EVENT_ERRORMESSAGE_CHANGED: {
+      // aria-errormessage was changed. If aria-invalid != "true", it means that
+      // VoiceOver should (a) expose a new message or (b) remove an
+      // old message
+      if (![[self moxInvalid] isEqualToString:@"false"]) {
+        [self moxPostNotification:@"AXValidationErrorChanged"];
+      }
+
+      break;
+    }
+  }
+}
+
+- (void)maybePostValidationErrorChanged {
+  NSArray* relations =
+      [self getRelationsByType:(mozilla::a11y::RelationType::ERRORMSG_FOR)];
+  if ([relations count] > 0) {
+    // only fire AXValidationErrorChanged if related node is not
+    // `aria-invalid="false"`
+    for (mozAccessible* related : relations) {
+      NSString* invalidStr = [related moxInvalid];
+      if (![invalidStr isEqualToString:@"false"]) {
+        [self moxPostNotification:@"AXValidationErrorChanged"];
+        break;
+      }
     }
   }
 }
