@@ -4,9 +4,8 @@ const { TorSettings, TorBridgeSource } = ChromeUtils.importESModule(
   "resource://gre/modules/TorSettings.sys.mjs"
 );
 
-const { TorConnect, TorConnectTopics } = ChromeUtils.importESModule(
-  "resource://gre/modules/TorConnect.sys.mjs"
-);
+const { TorConnect, TorConnectStage, TorConnectTopics } =
+  ChromeUtils.importESModule("resource://gre/modules/TorConnect.sys.mjs");
 
 const gBuiltinBridgeDialog = {
   init() {
@@ -79,14 +78,14 @@ const gBuiltinBridgeDialog = {
 
     this._acceptButton = dialog.getButton("accept");
 
-    Services.obs.addObserver(this, TorConnectTopics.StateChange);
+    Services.obs.addObserver(this, TorConnectTopics.StageChange);
 
     this.onSelectChange();
     this.onAcceptStateChange();
   },
 
   uninit() {
-    Services.obs.removeObserver(this, TorConnectTopics.StateChange);
+    Services.obs.removeObserver(this, TorConnectTopics.StageChange);
   },
 
   onSelectChange() {
@@ -96,18 +95,18 @@ const gBuiltinBridgeDialog = {
   },
 
   onAcceptStateChange() {
-    const connect = TorConnect.canBeginBootstrap;
+    const connect = TorConnect.stageName !== TorConnectStage.Bootstrapped;
     this._result.connect = connect;
     this._acceptButton.setAttribute(
       "data-l10n-id",
-      connect ? "bridge-dialog-button-connect" : "bridge-dialog-button-accept"
+      connect ? "bridge-dialog-button-connect2" : "bridge-dialog-button-accept2"
     );
     this._acceptButton.classList.toggle("tor-button", connect);
   },
 
   observe(subject, topic) {
     switch (topic) {
-      case TorConnectTopics.StateChange:
+      case TorConnectTopics.StageChange:
         this.onAcceptStateChange();
         break;
     }
